#include <stdio.h>
#include <pthread.h>
#include <net_tie.h>
#include <syslog.h>
#include <libcman.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>

#define DEFAULT_TOKEN 10000
#define DEFAULT_INTERVAL 1000
#define MIN_TOKEN 5000		/* Minimum token timeout (milliseconds) */
#define MIN_INTERVAL 250	/* Ping interval minimum (milliseconds) */


static int allow_soft = 0;
static int running = 1;

int debug = 0;

#define dbg_printf(level, fmt, args...) \
do { \
	if (debug >= level) \
		printf(fmt,##args); \
} while(0)


void
usage(char *name, int retval)
{
	printf("QNet - Simple network IPv4 Tiebreaker for CMAN\n\n");
	printf("usage: %s <host|address> [options]\n", name);
	printf(" -s       Make one node + IP tiebreaker sufficient to \n");
	printf("          form a quorum (DANGEROUS)\n");
	printf(" -f       Do not fork\n");
	printf(" -d	  Increase debug level (NOTE: implies -f)\n");
	printf(" -i <x>   Starting ping interval hint (milliseconds)\n");
	printf(" -t <x>   CoroSync / OpenAIS Token timeout (milliseconds)\n");
	printf("\n");
	printf("QNet only works with two node clusters with CMAN configured like so:\n");
	printf("  <cman expected_votes=\"2\" two_node=\"0\" />\n");
	printf("\n");
	exit(retval);
}

void
sigusr_handler(int sig)
{
	if (sig == SIGUSR1)
		allow_soft = 1;
	else
		allow_soft ^= 1;
}


void
exit_handler(int sig)
{
	running = 0;
}


int
node_count(cman_handle_t ch)
{
	cman_node_t *cman_nodes = NULL;
	int x = 0, retnodes = 0, ret = 0;

	x = cman_get_node_count(ch);
	if (x <= 0)
		return 0;

	cman_nodes = malloc(sizeof(cman_node_t) * x);
	if (!cman_nodes)
		return 0;

	if (cman_get_nodes(ch, x, &retnodes, cman_nodes) < 0) {
		free(cman_nodes);
		return 0;
	}

	for (x = 0; x < retnodes; x++) {
		if (cman_nodes[x].cn_nodeid > 0 && cman_nodes[x].cn_member)
			++ret;
	}

	free(cman_nodes);
	return ret;
}


int
main(int argc, char **argv)
{
	char *ip_addr = NULL;
	int op, daemonize = 1;
	int quorum = 0, count = 0, have_net, had_quorum = 0;
	int x, token = DEFAULT_TOKEN, interval = DEFAULT_INTERVAL, errors = 0;
	pthread_t thread;
	cman_handle_t ch;

	if (geteuid() != 0) {
		printf("You are not root.\n");
		return 1;
	}

	while ((op = getopt(argc, argv, "a:t:i:sfdh?")) != EOF) {
		switch(op) {
		case 't':
			token = atoi(optarg);
			if (token < MIN_TOKEN) {
				printf("Token value must be at least %dms",
				       MIN_TOKEN);
				errors++;
			}
			break;
		case 'i':
			interval = atoi(optarg);
			if (interval < MIN_INTERVAL) {
				printf("Ping interval must be at "
				       "least %dms\n", MIN_INTERVAL);
				errors++;
			}
			break;
		case 's':
			allow_soft = 1;
			break;
		case 'f':
			daemonize = 0;
			break;
		case 'd':
			daemonize = 0;
			debug++;
			break;
		case '?':
		case 'h':
			usage(argv[0], !!errors);
		}
	}

	/* Make sure there's an IP address on the command line */
	if (optind < argc)
		ip_addr = strdup(argv[optind]);
	if (!ip_addr)
		++errors;
	if (errors)
		usage(argv[0], 1);

	if (daemonize)
		daemon(0,0);

	do {
		ch = cman_admin_init(NULL);
		if (!ch)
			sleep(1);
	} while (!ch);

	signal(SIGINT, exit_handler);
	signal(SIGQUIT, exit_handler);
	signal(SIGTERM, exit_handler);
	signal(SIGUSR1, sigusr_handler);
	signal(SIGUSR2, sigusr_handler);

	net_tiebreaker_init(ip_addr, token * 1000, interval * 1000);
	net_create_quorum_thread(&thread);
	if (cman_register_quorum_device(ch, "QNet", 1) < 0) {
		// This still doesn't work ... need to talk to chrissie
		//printf("CMAN registration failed...!?\n");
		//exit(1);
	}

	while (running) {
		usleep(interval*1000);
		quorum = cman_is_quorate(ch);
		if (quorum)
			had_quorum = 1;
		count = node_count(ch);
		have_net = net_tiebreaker();

		dbg_printf(1, "q:%d hq:%d c:%d n:%d a:%d -> ",
			   quorum, had_quorum,
			   count, have_net, allow_soft);

		if (!quorum) {
			/* inquorate */
			/* Count the IP tiebreaker vote if:
			   (a) one member online,
			   (b) our ping IP is online, and
			   (c) the admin said 'force soft quorum' */
			if (have_net && count == 1 && allow_soft) {
				quorum = 1;
			}
		} else {
			/* Quorate */
			/* Don't count our IP tiebreaker if it isn't 
			   online */
			if (!have_net) {
				quorum = 0;
			}
		}

		dbg_printf(1, "%d\n", quorum);

		cman_poll_quorum_device(ch, quorum);
	}

	cman_poll_quorum_device(ch, 0);
	cman_unregister_quorum_device(ch);
	cman_finish(ch);
	net_cancel_quorum_thread();

	return 0;
}
